"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Observer = void 0;
const socket_io_client_1 = __importDefault(require("socket.io-client"));
const Constants_1 = require("./Constants");
const Entity_1 = require("./Entity");
const Player_1 = require("./Player");
const Tools_1 = require("./Tools");
class Observer {
    constructor(serverData, g) {
        this.entities = new Map();
        this.pingMap = new Map();
        this.pingNum = 1;
        this.pings = [];
        this.players = new Map();
        this.projectiles = new Map();
        this.S = {};
        this.serverData = serverData;
        this.G = g;
    }
    async connect(reconnect = false, start = true) {
        this.socket = socket_io_client_1.default(`ws://${this.serverData.addr}:${this.serverData.port}`, {
            autoConnect: false,
            reconnection: reconnect,
            transports: ["websocket"]
        });
        this.socket.on("action", (data) => {
            this.projectiles.set(data.pid, { ...data, date: new Date() });
        });
        this.socket.on("death", (data) => {
            this.deleteEntity(data.id);
        });
        this.socket.on("disappear", (data) => {
            this.players.delete(data.id) || this.entities.delete(data.id);
            this.updatePositions();
        });
        this.socket.on("entities", (data) => {
            this.parseEntities(data);
        });
        this.socket.on("hit", (data) => {
            if (data.miss || data.evade) {
                this.projectiles.delete(data.pid);
                return;
            }
            if (data.reflect) {
                const p = this.projectiles.get(data.pid);
                if (p) {
                    p.damage = data.reflect;
                    p.target = data.hid;
                    p.x = this.x;
                    p.y = this.y;
                }
            }
            if (data.kill) {
                this.projectiles.delete(data.pid);
                this.deleteEntity(data.id);
            }
            else if (data.damage) {
                this.projectiles.delete(data.pid);
                const e = this.entities.get(data.id);
                if (e)
                    e.hp = e.hp - data.damage;
            }
            else {
                this.projectiles.delete(data.pid);
            }
        });
        this.socket.on("new_map", (data) => {
            this.parseNewMap(data);
        });
        this.socket.on("ping_ack", (data) => {
            const ping = this.pingMap.get(data.id);
            if (ping) {
                const time = Date.now() - ping.time;
                this.pings.push(time);
                if (ping.log)
                    console.log(`Ping: ${time}`);
                if (this.pings.length > Constants_1.Constants.MAX_PINGS)
                    this.pings.shift();
                this.pingMap.delete(data.id);
            }
        });
        this.socket.on("server_info", (data) => {
            for (const mtype in data) {
                if (typeof data[mtype] !== "object")
                    continue;
                if (!data[mtype].live)
                    continue;
                const mN = mtype;
                const goodData = data[mN];
                if (goodData.hp == undefined) {
                    goodData.hp = this.G.monsters[mN].hp;
                    goodData.max_hp = this.G.monsters[mN].hp;
                }
            }
            this.S = data;
        });
        if (start) {
            console.debug(`Connecting to ${this.serverData.region}${this.serverData.name}...`);
            const connected = new Promise((resolve, reject) => {
                this.socket.on("welcome", (data) => {
                    if (data.region !== this.serverData.region || data.name !== this.serverData.name) {
                        reject(`We wanted the server ${this.serverData.region}${this.serverData.name}, but we are on ${data.region}${data.name}.`);
                    }
                    else {
                        this.socket.emit("loaded", {
                            height: 1080,
                            scale: 2,
                            success: 1,
                            width: 1920
                        });
                        resolve();
                    }
                });
                setTimeout(() => {
                    reject(`Failed to start within ${Constants_1.Constants.CONNECT_TIMEOUT_MS / 1000}s.`);
                }, Constants_1.Constants.CONNECT_TIMEOUT_MS);
            });
            this.socket.open();
            return connected;
        }
    }
    async deleteEntity(id) {
        const entity = this.entities.get(id);
        if (!entity)
            return;
        if (this.S[entity.type])
            delete this.S[entity.type];
        this.entities.delete(id);
    }
    async parseEntities(data) {
        if (data.type == "all") {
            this.lastAllEntities = Date.now();
            this.entities.clear();
            this.players.clear();
            this.lastPositionUpdate = Date.now();
        }
        else {
            this.updatePositions();
        }
        for (const monster of data.monsters) {
            if (!this.entities.has(monster.id)) {
                const e = new Entity_1.Entity(monster, data.map, this.G);
                this.entities.set(monster.id, e);
            }
            else {
                const e = this.entities.get(monster.id);
                e.updateData(monster);
            }
        }
        for (const player of data.players) {
            if (!this.players.has(player.id)) {
                const p = new Player_1.Player(player, data.map, this.G);
                this.players.set(player.id, p);
            }
            else {
                const p = this.players.get(player.id);
                p.updateData(player);
            }
        }
    }
    parseNewMap(data) {
        this.projectiles.clear();
        this.x = data.x;
        this.y = data.y;
        this.map = data.name;
        this.parseEntities(data.entities);
    }
    updatePositions() {
        if (this.lastPositionUpdate) {
            const msSinceLastUpdate = Date.now() - this.lastPositionUpdate;
            if (msSinceLastUpdate == 0)
                return;
            for (const [, entity] of this.entities) {
                if (!entity.moving)
                    continue;
                const speed = entity.speed;
                const distanceTravelled = speed * msSinceLastUpdate / 1000;
                const angle = Math.atan2(entity.going_y - entity.y, entity.going_x - entity.x);
                const distanceToGoal = Tools_1.Tools.distance({ x: entity.x, y: entity.y }, { x: entity.going_x, y: entity.going_y });
                if (distanceTravelled > distanceToGoal) {
                    entity.moving = false;
                    entity.x = entity.going_x;
                    entity.y = entity.going_y;
                }
                else {
                    entity.x = entity.x + Math.cos(angle) * distanceTravelled;
                    entity.y = entity.y + Math.sin(angle) * distanceTravelled;
                }
                for (const condition in entity.s) {
                    const newCooldown = entity.s[condition].ms - msSinceLastUpdate;
                    if (newCooldown <= 0)
                        delete entity.s[condition];
                    else
                        entity.s[condition].ms = newCooldown;
                }
            }
            for (const [, player] of this.players) {
                if (!player.moving)
                    continue;
                const distanceTravelled = player.speed * msSinceLastUpdate / 1000;
                const angle = Math.atan2(player.going_y - player.y, player.going_x - player.x);
                const distanceToGoal = Tools_1.Tools.distance({ x: player.x, y: player.y }, { x: player.going_x, y: player.going_y });
                if (distanceTravelled > distanceToGoal) {
                    player.moving = false;
                    player.x = player.going_x;
                    player.y = player.going_y;
                }
                else {
                    player.x = player.x + Math.cos(angle) * distanceTravelled;
                    player.y = player.y + Math.sin(angle) * distanceTravelled;
                }
                for (const condition in player.s) {
                    const newCooldown = player.s[condition].ms - msSinceLastUpdate;
                    if (newCooldown <= 0)
                        delete player.s[condition];
                    else
                        player.s[condition].ms = newCooldown;
                }
            }
        }
        let toDelete = [];
        for (const [id, entity] of this.entities) {
            if (Tools_1.Tools.distance(this, entity) < Constants_1.Constants.MAX_VISIBLE_RANGE)
                continue;
            toDelete.push(id);
        }
        for (const id of toDelete)
            this.entities.delete(id);
        toDelete = [];
        for (const [id, player] of this.players) {
            if (Tools_1.Tools.distance(this, player) < Constants_1.Constants.MAX_VISIBLE_RANGE)
                continue;
            toDelete.push(id);
        }
        for (const id of toDelete)
            this.players.delete(id);
        for (const [id, projectile] of this.projectiles) {
            if (Date.now() - projectile.date.getTime() > Constants_1.Constants.STALE_PROJECTILE_MS)
                this.projectiles.delete(id);
        }
        this.lastPositionUpdate = Date.now();
    }
    sendPing(log = true) {
        const pingID = this.pingNum.toString();
        this.pingNum++;
        this.pingMap.set(pingID, { log: log, time: Date.now() });
        this.socket.emit("ping_trig", { id: pingID });
        return pingID;
    }
    getNearestMonster(mtype) {
        let closest;
        let closestD = Number.MAX_VALUE;
        for (const [, entity] of this.entities) {
            if (mtype && entity.type !== mtype)
                continue;
            const d = Tools_1.Tools.distance(this, entity);
            if (d < closestD) {
                closest = entity;
                closestD = d;
            }
        }
        if (closest)
            return { distance: closestD, monster: closest };
    }
}
exports.Observer = Observer;
//# sourceMappingURL=Observer.js.map